@@ -9,15 +9,15 @@ App::Ack - A container for functions for the ack program
 
 =head1 VERSION
 
-Version 1.75_01
+Version 1.77_01
 
 =cut
 
 our $VERSION;
 our $COPYRIGHT;
 BEGIN {
-    $VERSION = '1.75_01';
-    $COPYRIGHT = 'Copyright 2005-2007 Andy Lester, all rights reserved.';
+    $VERSION = '1.77_01';
+    $COPYRIGHT = 'Copyright 2005-2008 Andy Lester, all rights reserved.';
 }
 
 our %types;
@@ -54,6 +54,7 @@ BEGIN {
         _sgbak              => 'Vault/Fortress',
         'autom4te.cache'    => 'autoconf',
         'cover_db'          => 'Devel::Cover',
+        _build              => 'Module::Build',
     );
 
     %mappings = (
@@ -68,22 +69,25 @@ BEGIN {
         fortran     => [qw( f f77 f90 f95 f03 for ftn fpp )],
         haskell     => [qw( hs lhs )],
         hh          => [qw( h )],
-        html        => [qw( htm html shtml )],
-        skipped     => q{Files, but not directories, normally skipped by ack (default: off)},
+        html        => [qw( htm html shtml xhtml )],
         lisp        => [qw( lisp )],
         java        => [qw( java properties )],
         js          => [qw( js )],
         jsp         => [qw( jsp jspx jhtm jhtml )],
         make        => q{Makefiles},
         mason       => [qw( mas mhtml mpl mtxt )],
+        objc        => [qw( m h )],
+        objcpp      => [qw( mm h )],
         ocaml       => [qw( ml mli )],
         parrot      => [qw( pir pasm pmc ops pod pg tg )],
         perl        => [qw( pl pm pod t )],
         php         => [qw( php phpt php3 php4 php5 )],
+        plone       => [qw( pt cpt metadata cpy py )],
         python      => [qw( py )],
         ruby        => [qw( rb rhtml rjs rxml )],
         scheme      => [qw( scm )],
         shell       => [qw( sh bash csh ksh zsh )],
+        skipped     => q{Files, but not directories, normally skipped by ack (default: off)},
         sql         => [qw( sql ctl )],
         tcl         => [qw( tcl )],
         tex         => [qw( tex cls sty )],
@@ -181,6 +185,7 @@ sub get_command_line_options {
         o                       => sub { $opt{output} = '$&' },
         'output=s'              => \$opt{output},
         'passthru'              => \$opt{passthru},
+        'print0'                => \$opt{print0},
         'Q|literal'             => \$opt{Q},
         'sort-files'            => \$opt{sort_files},
         'u|unrestricted'        => \$opt{u},
@@ -191,7 +196,7 @@ sub get_command_line_options {
         'version'   => sub { print_version_statement(); exit 1; },
         'help|?:s'  => sub { shift; show_help(@_); exit; },
         'help-types'=> sub { show_help_types(); exit; },
-        'man'       => sub {require Pod::Usage; Pod::Usage::pod2usage({-verbose => 2}); exit; },
+        'man'       => sub { require Pod::Usage; Pod::Usage::pod2usage({-verbose => 2}); exit; },
 
         'type=s'    => sub {
             # Whatever --type=xxx they specify, set it manually in the hash
@@ -208,17 +213,24 @@ sub get_command_line_options {
         }, # type sub
     };
 
-    for my $i ( filetypes_supported() ) {
-        $getopt_specs->{ "$i!" } = \$type_wanted{ $i };
-    }
-
     # Stick any default switches at the beginning, so they can be overridden
     # by the command line switches.
     unshift @ARGV, split( ' ', $ENV{ACK_OPTIONS} ) if defined $ENV{ACK_OPTIONS};
 
-    Getopt::Long::Configure( 'bundling', 'no_ignore_case' );
-    Getopt::Long::GetOptions( %{$getopt_specs} ) or
-        App::Ack::die( 'See ack --help or ack --man for options.' );
+    # first pass through options, looking for type definitions
+    def_types_from_ARGV();
+
+    for my $i ( filetypes_supported() ) {
+        $getopt_specs->{ "$i!" } = \$type_wanted{ $i };
+    }
+
+
+    {
+        my $parser = Getopt::Long::Parser->new();
+        $parser->configure( 'bundling', 'no_ignore_case', );
+        $parser->getoptions( %{$getopt_specs} ) or
+            App::Ack::die( 'See ack --help or ack --man for options.' );
+    }
 
     my %defaults = (
         all            => 0,
@@ -284,6 +296,84 @@ sub get_command_line_options {
     return %opt;
 }
 
+=head2 def_types_from_ARGV
+
+Go through the command line arguments and look for
+I<--type-set foo=.foo,.bar> and I<--type-add xml=.rdf>.
+Remove them from @ARGV and add them to the supported filetypes,
+i.e. into %mappings, etc.
+
+=cut
+
+sub def_types_from_ARGV {
+    my @typedef;
+
+    my $parser = Getopt::Long::Parser->new();
+        # pass_through   => leave unrecognized command line arguments alone
+        # no_auto_abbrev => otherwise -c is expanded and not left alone
+    $parser->configure( 'no_ignore_case', 'pass_through', 'no_auto_abbrev' );
+    $parser->getoptions( 
+        'type-set=s' => sub { shift; push @typedef, ['c', shift] },
+        'type-add=s' => sub { shift; push @typedef, ['a', shift] },
+    ) or App::Ack::die( 'See ack --help or ack --man for options.' );
+
+    for my $td (@typedef) {
+        my ($type, $ext) = split '=', $td->[1];
+
+        if ( $td->[0] eq 'c' ) {
+            # type-set
+
+            # can't redefine types 'make', 'skipped', 'text' and 'binary'
+            App::Ack::die( "--type-set: Builtin type '$type' cannot be changed." )
+                if exists $mappings{$type} && ref $mappings{$type} ne 'ARRAY';
+
+            delete_type($type) if exists $mappings{$type};
+        } else {
+            # type-add
+
+            # can't append to types 'make', 'skipped', 'text' and 'binary'
+            App::Ack::die( "--type-add: Builtin type '$type' cannot be changed." )
+                if exists $mappings{$type} && ref $mappings{$type} ne 'ARRAY';
+
+            App::Ack::warn( "--type-add: Type '$type' does not exist, creating with '$ext' ..." )
+                unless exists $mappings{$type};
+        }
+
+        my @exts = map { s/^\.//; $_ } split ',', $ext; # %types stores extensions without leading '.'
+
+        if ( !exists $mappings{$type} || ref($mappings{$type}) eq 'ARRAY' ) {
+            push @{$mappings{$type}}, @exts;
+            for my $e ( @exts ) {
+                push @{$types{$e}}, $type;
+            }
+        } else {
+            App::Ack::die( "Cannot append to type '$type'." );
+        }
+    }
+
+    return;
+}
+
+=head2 delete_type
+
+Removes a type from the internal structures containing type
+information: %mappings, %types and %type_wanted.
+
+=cut
+
+sub delete_type {
+    my $type = shift;
+
+    App::Ack::die( "Internal error: Cannot delete builtin type '$type'." )
+        unless ref $mappings{$type} eq 'ARRAY';
+
+    delete $mappings{$type};
+    delete $type_wanted{$type};
+    for my $ext ( keys %types ) {
+        $types{$ext} = [ grep { $_ ne $type } @{$types{$ext}} ];
+    }
+}
+
 =head2 skipdir_filter
 
 Standard filter to pass as a L<File::Next> descend_filter.  It
@@ -380,6 +470,7 @@ sub is_searchable {
     my $filename = shift;
 
     # If these are updated, update the --help message
+    return if $filename =~ /\.bak$/;
     return if $filename =~ /~$/;
     return if $filename =~ m{$path_sep_regex?(?:#.+#|core\.\d+|[._].*\.swp)$}o;
 
@@ -402,7 +493,7 @@ sub build_regex {
         $str = "$str\\b" if $str =~ /\w$/;
     }
 
-    return $opt->{i} ? qr/$str/i : qr/$str/;
+    return $str;
 }
 
 
@@ -526,6 +617,9 @@ Search output:
   -C [NUM], --context[=NUM]
                         Print NUM lines (default 2) of output context.
 
+  --print0              Print null byte as separator between filenames,
+                        only works with -f, -g, -l, -L or -c.
+
 File finding:
   -f                    Only print the files found, without searching.
                         The PATTERN must not be specified.
@@ -541,6 +635,15 @@ File inclusion/exclusion:
   --noperl              Exclude Perl files.
   --type=noperl         Exclude Perl files.
                         See "ack --help type" for supported filetypes.
+
+  --type-add TYPE=.EXTENSION[,.EXT2[,...]]
+                        Files with the given EXTENSION(s) are recognized as
+                        being of (the existing) type TYPE
+  --type-set TYPE=.EXTENSION[,.EXT2[,...]]
+                        Files with the given EXTENSION(s) are recognized as
+                        being of type TYPE. This replaces an existing
+                        definition for type TYPE. 
+
   --[no]follow          Follow symlinks.  Default is off.
 
   Directories ignored by default:
@@ -738,7 +841,7 @@ sub close_file {
 }
 
 
-=head2 needs_line_scan( $fh, $regex )
+=head2 needs_line_scan( $fh, $regex, \%opts )
 
 Slurp up an entire file up to 100K, see if there are any matches
 in it, and if so, let us know so we can iterate over it directly.
@@ -749,6 +852,7 @@ If it's bigger than 100K, we have to do the line-by-line, too.
 sub needs_line_scan {
     my $fh = shift;
     my $regex = shift;
+    my $opt = shift;
 
     my $size = -s $fh;
 
@@ -758,16 +862,14 @@ sub needs_line_scan {
 
     my $buffer;
     my $rc = sysread( $fh, $buffer, $size );
-    if ( defined($rc) && ( $rc == $size ) && ( $buffer =~ /$regex/osm ) ) {
-        return 1;
-    }
-    else {
-        return 0;
-    }
+    return 0 unless $rc && ( $rc == $size );
+
+    $regex = $opt->{i} ? qr/$regex/im : qr/$regex/m;
+    return ( $buffer =~ /$regex/ );
 }
 
 
-=head2 search( $fh, $could_be_binary, $filename, $regex, \%opt )
+=head2 search( $fh, $could_be_binary, $filename, \%opt )
 
 Main search method
 
@@ -788,7 +890,6 @@ sub search {
     my $fh = shift;
     my $could_be_binary = shift;
     $filename = shift;
-    $regex = shift;
     my $opt = shift;
 
     my $v = $opt->{v};
@@ -806,6 +907,10 @@ sub search {
         @lines = ( @{$opt->{lines}}, -1 );
         undef $regex; # Don't match when printing matching line
     }
+    else {
+        $regex = $opt->{i} ? qr/$opt->{regex}/i : qr/$opt->{regex}/;
+    }
+
 
     # for context processing
     $last_output_line = -1;
@@ -939,8 +1044,9 @@ sub print_match_or_context {
         }
         else {
             if ( $color && $is_match && $regex ) {
-                if ( s/($regex)/Term::ANSIColor::colored($1,$ENV{ACK_COLOR_MATCH})/eg ) {
-                    s/\n$/\e[0m\e[K\n/;     # Before \n, reset the color and clear to end of line
+                if ( s/$regex/Term::ANSIColor::colored( substr($_, $-[0], $+[0] - $-[0]), $ENV{ACK_COLOR_MATCH} )/eg ) {
+                    # At the end of the line reset the color and keep existing line ending
+                    s/([\r\n]*)$/\e[0m\e[K$1/;
                 }
             }
             print;
@@ -955,7 +1061,7 @@ sub print_match_or_context {
 } # scope around search() and print_match_or_context()
 
 
-=head2 search_and_list( $fh, $filename, $regex, \%opt )
+=head2 search_and_list( $fh, $filename, \%opt )
 
 Optimized version of searching for -l and --count, which do not
 show lines.
@@ -965,11 +1071,13 @@ show lines.
 sub search_and_list {
     my $fh = shift;
     my $filename = shift;
-    my $regex = shift;
     my $opt = shift;
 
     my $nmatches = 0;
     my $count = $opt->{count};
+    my $ors = $opt->{print0} ? "\0" : "\n"; # output record separator
+
+    my $regex = $opt->{i} ? qr/$opt->{regex}/i : qr/$opt->{regex}/;
 
     if ( $opt->{v} ) {
         while (<$fh>) {
@@ -993,10 +1101,10 @@ sub search_and_list {
     if ( $nmatches ) {
         print $filename;
         print ':', $nmatches if $count;
-        print "\n";
+        print $ors;
     }
     elsif ( $count && !$opt->{l} ) {
-        print "$filename:0\n";
+        print "$filename:0", $ors;
     }
 
     return $nmatches ? 1 : 0;
@@ -1014,22 +1122,30 @@ sub filetypes_supported_set {
 }
 
 
-=head2 print_files( $iter, $one [, $regex] )
+=head2 print_files( $iter, $one [, $regex, [, $ors ]] )
 
 Prints all the files returned by the iterator matching I<$regex>.
+
 If I<$one> is set, stop after the first.
+The output record separator I<$ors> defaults to C<"\n"> and defines, what to
+print after each filename.
 
 =cut
 
 sub print_files {
     my $iter = shift;
-    my $one = shift;
-    my $regex = shift;
+    my $opt = shift;
+
+    my $regex;
+    if ( $opt->{g} ) {
+        $regex = $opt->{i} ? qr/$opt->{g}/i : qr/$opt->{g}/;
+    }
+    my $ors = $opt->{print0} ? "\0" : "\n";
 
     while ( defined ( my $file = $iter->() ) ) {
         if ( (not defined $regex) || ($file =~ m/$regex/o) ) {
-            print $file, "\n";
-            last if $one;
+            print $file, $ors;
+            last if $opt->{1};
         }
     }
 
